@@ -28,7 +28,7 @@ import re
 import time
 
 from contextlib import contextmanager
-from distutils.version import LooseVersion
+from ansible.module_utils.compat.version import LooseVersion
 from numbers import Number
 from traceback import format_exc
 
@@ -42,7 +42,15 @@ from jinja2.loaders import FileSystemLoader
 from jinja2.runtime import Context, StrictUndefined
 
 from ansible import constants as C
-from ansible.errors import AnsibleError, AnsibleFilterError, AnsiblePluginRemovedError, AnsibleUndefinedVariable, AnsibleAssertionError
+from ansible.errors import (
+    AnsibleAssertionError,
+    AnsibleError,
+    AnsibleFilterError,
+    AnsibleLookupError,
+    AnsibleOptionsError,
+    AnsiblePluginRemovedError,
+    AnsibleUndefinedVariable,
+)
 from ansible.module_utils.six import iteritems, string_types, text_type
 from ansible.module_utils.six.moves import range
 from ansible.module_utils._text import to_native, to_text, to_bytes
@@ -87,10 +95,11 @@ if C.DEFAULT_JINJA2_NATIVE:
     except ImportError:
         from jinja2 import Environment
         from jinja2.utils import concat as j2_concat
-        display.warning(
-            'jinja2_native requires Jinja 2.10 and above. '
-            'Version detected: %s. Falling back to default.' % j2_version
-        )
+        if C.JINJA2_NATIVE_WARNING:
+            display.warning(
+                'jinja2_native requires Jinja 2.10 and above. '
+                'Version detected: %s. Falling back to default.' % j2_version
+            )
 
 
 JINJA2_BEGIN_TOKENS = frozenset(('variable_begin', 'block_begin', 'comment_begin', 'raw_begin'))
@@ -180,6 +189,25 @@ def _escape_backslashes(data, jinja_env):
     return data
 
 
+def is_possibly_template(data, jinja_env):
+    """Determines if a string looks like a template, by seeing if it
+    contains a jinja2 start delimiter. Does not guarantee that the string
+    is actually a template.
+
+    This is different than ``is_template`` which is more strict.
+    This method may return ``True`` on a string that is not templatable.
+
+    Useful when guarding passing a string for templating, but when
+    you want to allow the templating engine to make the final
+    assessment which may result in ``TemplateSyntaxError``.
+    """
+    if isinstance(data, string_types):
+        for marker in (jinja_env.block_start_string, jinja_env.variable_start_string, jinja_env.comment_start_string):
+            if marker in data:
+                return True
+    return False
+
+
 def is_template(data, jinja_env):
     """This function attempts to quickly detect whether a value is a jinja2
     template. To do so, we look for the first 2 matching jinja2 tokens for
@@ -190,6 +218,11 @@ def is_template(data, jinja_env):
     comment = False
     d2 = jinja_env.preprocess(data)
 
+    # Quick check to see if this is remotely like a template before doing
+    # more expensive investigation.
+    if not is_possibly_template(d2, jinja_env):
+        return False
+
     # This wraps a lot of code, but this is due to lex returning a generator
     # so we may get an exception at any part of the loop
     try:
@@ -430,8 +463,12 @@ class JinjaPluginIntercept(MutableMapping):
             return
 
         for plugin in self._pluginloader.all():
-            method_map = getattr(plugin, self._method_map_name)
-            self._delegatee.update(method_map())
+            try:
+                method_map = getattr(plugin, self._method_map_name)
+                self._delegatee.update(method_map())
+            except Exception as e:
+                display.warning("Skipping %s plugin %s as it seems to be invalid: %r" % (self._dirname, to_text(plugin._original_path), e))
+                continue
 
         if self._pluginloader.class_name == 'FilterModule':
             for plugin_name, plugin in self._delegatee.items():
@@ -536,9 +573,16 @@ class JinjaPluginIntercept(MutableMapping):
                 except Exception as e:
                     raise TemplateSyntaxError(to_native(e), 0)
 
-                method_map = getattr(plugin_impl, self._method_map_name)
+                try:
+                    method_map = getattr(plugin_impl, self._method_map_name)
+                    func_items = iteritems(method_map())
+                except Exception as e:
+                    display.warning(
+                        "Skipping %s plugin %s as it seems to be invalid: %r" % (self._dirname, to_text(plugin_impl._original_path), e),
+                    )
+                    continue
 
-                for func_name, func in iteritems(method_map()):
+                for func_name, func in func_items:
                     fq_name = '.'.join((parent_prefix, func_name))
                     # FIXME: detect/warn on intra-collection function name collisions
                     if self._pluginloader.class_name == 'FilterModule':
@@ -628,10 +672,6 @@ class Templar:
         self._cached_result = {}
         self._basedir = loader.get_basedir() if loader else './'
 
-        # flags to determine whether certain failures during templating
-        # should result in fatal errors being raised
-        self._fail_on_lookup_errors = True
-        self._fail_on_filter_errors = True
         self._fail_on_undefined_errors = C.DEFAULT_UNDEFINED_VAR_BEHAVIOR
 
         environment_class = AnsibleNativeEnvironment if USE_JINJA2_NATIVE else AnsibleEnvironment
@@ -652,6 +692,7 @@ class Templar:
         self.environment.globals['query'] = self.environment.globals['q'] = self._query_lookup
         self.environment.globals['now'] = self._now_datetime
         self.environment.globals['finalize'] = self._finalize
+        self.environment.globals['undef'] = self._make_undefined
 
         # the current rendering context under which the templar class is working
         self.cur_context = None
@@ -683,6 +724,7 @@ class Templar:
         new_templar = object.__new__(Templar)
         new_templar.__dict__.update(self.__dict__)
         new_templar.environment = new_env
+        new_templar.environment.finalize = new_templar._finalize
 
         mapping = {
             'available_variables': new_templar,
@@ -787,107 +829,99 @@ class Templar:
         if fail_on_undefined is None:
             fail_on_undefined = self._fail_on_undefined_errors
 
-        try:
-            if convert_bare:
-                variable = self._convert_bare_variable(variable)
+        if convert_bare:
+            variable = self._convert_bare_variable(variable)
 
-            if isinstance(variable, string_types):
-                result = variable
-
-                if self.is_possibly_template(variable):
-                    # Check to see if the string we are trying to render is just referencing a single
-                    # var.  In this case we don't want to accidentally change the type of the variable
-                    # to a string by using the jinja template renderer. We just want to pass it.
-                    only_one = self.SINGLE_VAR.match(variable)
-                    if only_one:
-                        var_name = only_one.group(1)
-                        if var_name in self._available_variables:
-                            resolved_val = self._available_variables[var_name]
-                            if isinstance(resolved_val, NON_TEMPLATED_TYPES):
-                                return resolved_val
-                            elif resolved_val is None:
-                                return C.DEFAULT_NULL_REPRESENTATION
-
-                    # Using a cache in order to prevent template calls with already templated variables
-                    sha1_hash = None
-                    if cache:
-                        variable_hash = sha1(text_type(variable).encode('utf-8'))
-                        options_hash = sha1(
-                            (
-                                text_type(preserve_trailing_newlines) +
-                                text_type(escape_backslashes) +
-                                text_type(fail_on_undefined) +
-                                text_type(overrides)
-                            ).encode('utf-8')
-                        )
-                        sha1_hash = variable_hash.hexdigest() + options_hash.hexdigest()
-                    if cache and sha1_hash in self._cached_result:
-                        result = self._cached_result[sha1_hash]
-                    else:
-                        result = self.do_template(
-                            variable,
-                            preserve_trailing_newlines=preserve_trailing_newlines,
-                            escape_backslashes=escape_backslashes,
-                            fail_on_undefined=fail_on_undefined,
-                            overrides=overrides,
-                            disable_lookups=disable_lookups,
-                        )
-
-                        if not self.jinja2_native:
-                            unsafe = hasattr(result, '__UNSAFE__')
-                            if convert_data and not self._no_type_regex.match(variable):
-                                # if this looks like a dictionary or list, convert it to such using the safe_eval method
-                                if (result.startswith("{") and not result.startswith(self.environment.variable_start_string)) or \
-                                        result.startswith("[") or result in ("True", "False"):
-                                    eval_results = safe_eval(result, include_exceptions=True)
-                                    if eval_results[1] is None:
-                                        result = eval_results[0]
-                                        if unsafe:
-                                            result = wrap_var(result)
-                                    else:
-                                        # FIXME: if the safe_eval raised an error, should we do something with it?
-                                        pass
-
-                        # we only cache in the case where we have a single variable
-                        # name, to make sure we're not putting things which may otherwise
-                        # be dynamic in the cache (filters, lookups, etc.)
-                        if cache and only_one:
-                            self._cached_result[sha1_hash] = result
-
-                return result
-
-            elif is_sequence(variable):
-                return [self.template(
-                    v,
-                    preserve_trailing_newlines=preserve_trailing_newlines,
-                    fail_on_undefined=fail_on_undefined,
-                    overrides=overrides,
-                    disable_lookups=disable_lookups,
-                ) for v in variable]
-            elif isinstance(variable, Mapping):
-                d = {}
-                # we don't use iteritems() here to avoid problems if the underlying dict
-                # changes sizes due to the templating, which can happen with hostvars
-                for k in variable.keys():
-                    if k not in static_vars:
-                        d[k] = self.template(
-                            variable[k],
-                            preserve_trailing_newlines=preserve_trailing_newlines,
-                            fail_on_undefined=fail_on_undefined,
-                            overrides=overrides,
-                            disable_lookups=disable_lookups,
-                        )
-                    else:
-                        d[k] = variable[k]
-                return d
-            else:
+        if isinstance(variable, string_types):
+            if not self.is_possibly_template(variable):
                 return variable
 
-        except AnsibleFilterError:
-            if self._fail_on_filter_errors:
-                raise
-            else:
-                return variable
+            # Check to see if the string we are trying to render is just referencing a single
+            # var.  In this case we don't want to accidentally change the type of the variable
+            # to a string by using the jinja template renderer. We just want to pass it.
+            only_one = self.SINGLE_VAR.match(variable)
+            if only_one:
+                var_name = only_one.group(1)
+                if var_name in self._available_variables:
+                    resolved_val = self._available_variables[var_name]
+                    if isinstance(resolved_val, NON_TEMPLATED_TYPES):
+                        return resolved_val
+                    elif resolved_val is None:
+                        return C.DEFAULT_NULL_REPRESENTATION
+
+            # Using a cache in order to prevent template calls with already templated variables
+            sha1_hash = None
+            if cache:
+                variable_hash = sha1(text_type(variable).encode('utf-8'))
+                options_hash = sha1(
+                    (
+                        text_type(preserve_trailing_newlines) +
+                        text_type(escape_backslashes) +
+                        text_type(fail_on_undefined) +
+                        text_type(overrides)
+                    ).encode('utf-8')
+                )
+                sha1_hash = variable_hash.hexdigest() + options_hash.hexdigest()
+
+                if sha1_hash in self._cached_result:
+                    return self._cached_result[sha1_hash]
+
+            result = self.do_template(
+                variable,
+                preserve_trailing_newlines=preserve_trailing_newlines,
+                escape_backslashes=escape_backslashes,
+                fail_on_undefined=fail_on_undefined,
+                overrides=overrides,
+                disable_lookups=disable_lookups,
+            )
+
+            if not self.jinja2_native:
+                unsafe = hasattr(result, '__UNSAFE__')
+                if convert_data and not self._no_type_regex.match(variable):
+                    # if this looks like a dictionary or list, convert it to such using the safe_eval method
+                    if (result.startswith("{") and not result.startswith(self.environment.variable_start_string)) or \
+                            result.startswith("[") or result in ("True", "False"):
+                        eval_results = safe_eval(result, include_exceptions=True)
+                        if eval_results[1] is None:
+                            result = eval_results[0]
+                            if unsafe:
+                                result = wrap_var(result)
+                        # FIXME: if the safe_eval raised an error, should we do something with it?
+
+            # we only cache in the case where we have a single variable
+            # name, to make sure we're not putting things which may otherwise
+            # be dynamic in the cache (filters, lookups, etc.)
+            if cache and only_one:
+                self._cached_result[sha1_hash] = result
+
+            return result
+
+        elif is_sequence(variable):
+            return [self.template(
+                v,
+                preserve_trailing_newlines=preserve_trailing_newlines,
+                fail_on_undefined=fail_on_undefined,
+                overrides=overrides,
+                disable_lookups=disable_lookups,
+            ) for v in variable]
+        elif isinstance(variable, Mapping):
+            d = {}
+            # we don't use iteritems() here to avoid problems if the underlying dict
+            # changes sizes due to the templating, which can happen with hostvars
+            for k in variable.keys():
+                if k not in static_vars:
+                    d[k] = self.template(
+                        variable[k],
+                        preserve_trailing_newlines=preserve_trailing_newlines,
+                        fail_on_undefined=fail_on_undefined,
+                        overrides=overrides,
+                        disable_lookups=disable_lookups,
+                    )
+                else:
+                    d[k] = variable[k]
+            return d
+        else:
+            return variable
 
     def is_template(self, data):
         '''lets us know if data has a template'''
@@ -906,23 +940,7 @@ class Templar:
     templatable = is_template
 
     def is_possibly_template(self, data):
-        '''Determines if a string looks like a template, by seeing if it
-        contains a jinja2 start delimiter. Does not guarantee that the string
-        is actually a template.
-
-        This is different than ``is_template`` which is more strict.
-        This method may return ``True`` on a string that is not templatable.
-
-        Useful when guarding passing a string for templating, but when
-        you want to allow the templating engine to make the final
-        assessment which may result in ``TemplateSyntaxError``.
-        '''
-        env = self.environment
-        if isinstance(data, string_types):
-            for marker in (env.block_start_string, env.variable_start_string, env.comment_start_string):
-                if marker in data:
-                    return True
-        return False
+        return is_possibly_template(data, self.environment)
 
     def _convert_bare_variable(self, variable):
         '''
@@ -997,16 +1015,30 @@ class Templar:
             ran = instance.run(loop_terms, variables=self._available_variables, **kwargs)
         except (AnsibleUndefinedVariable, UndefinedError) as e:
             raise AnsibleUndefinedVariable(e)
+        except AnsibleOptionsError as e:
+            # invalid options given to lookup, just reraise
+            raise e
+        except AnsibleLookupError as e:
+            # lookup handled error but still decided to bail
+            msg = 'Lookup failed but the error is being ignored: %s' % to_native(e)
+            if errors == 'warn':
+                display.warning(msg)
+            elif errors == 'ignore':
+                display.display(msg, log_only=True)
+            else:
+                raise e
+            return [] if wantlist else None
         except Exception as e:
-            if self._fail_on_lookup_errors:
-                msg = u"An unhandled exception occurred while running the lookup plugin '%s'. Error was a %s, original message: %s" % \
-                      (name, type(e), to_text(e))
-                if errors == 'warn':
-                    display.warning(msg)
-                elif errors == 'ignore':
-                    display.display(msg, log_only=True)
-                else:
-                    raise AnsibleError(to_native(msg))
+            # errors not handled by lookup
+            msg = u"An unhandled exception occurred while running the lookup plugin '%s'. Error was a %s, original message: %s" % \
+                  (name, type(e), to_text(e))
+            if errors == 'warn':
+                display.warning(msg)
+            elif errors == 'ignore':
+                display.display(msg, log_only=True)
+            else:
+                display.vvv('exception during Jinja2 execution: {0}'.format(format_exc()))
+                raise AnsibleError(to_native(msg), orig_exc=e)
             return [] if wantlist else None
 
         if ran and allow_unsafe is False:
@@ -1037,6 +1069,13 @@ class Templar:
 
         return ran
 
+    def _make_undefined(self, hint=None):
+        from jinja2.runtime import Undefined
+
+        if hint is None or isinstance(hint, Undefined) or hint == '':
+            hint = "Mandatory variable has not been overridden"
+        return AnsibleUndefined(hint)
+
     def do_template(self, data, preserve_trailing_newlines=True, escape_backslashes=True, fail_on_undefined=None, overrides=None, disable_lookups=False):
         if self.jinja2_native and not isinstance(data, string_types):
             return data
@@ -1048,15 +1087,21 @@ class Templar:
         if fail_on_undefined is None:
             fail_on_undefined = self._fail_on_undefined_errors
 
+        has_template_overrides = data.startswith(JINJA2_OVERRIDE)
+
         try:
-            # allows template header overrides to change jinja2 options.
-            if overrides is None:
-                myenv = self.environment
-            else:
+            # NOTE Creating an overlay that lives only inside do_template means that overrides are not applied
+            # when templating nested variables in AnsibleJ2Vars where Templar.environment is used, not the overlay.
+            # This is historic behavior that is kept for backwards compatibility.
+            if overrides:
                 myenv = self.environment.overlay(overrides)
+            elif has_template_overrides:
+                myenv = self.environment.overlay()
+            else:
+                myenv = self.environment
 
             # Get jinja env overrides from template
-            if hasattr(data, 'startswith') and data.startswith(JINJA2_OVERRIDE):
+            if has_template_overrides:
                 eol = data.find('\n')
                 line = data[len(JINJA2_OVERRIDE):eol]
                 data = data[eol + 1:]
@@ -1092,7 +1137,8 @@ class Templar:
                     res = ansible_native_concat(rf)
                 else:
                     res = j2_concat(rf)
-                if getattr(new_context, 'unsafe', False):
+                unsafe = getattr(new_context, 'unsafe', False)
+                if unsafe:
                     res = wrap_var(res)
             except TypeError as te:
                 if 'AnsibleUndefined' in to_native(te):
@@ -1122,6 +1168,8 @@ class Templar:
                 res_newlines = _count_newlines_from_end(res)
                 if data_newlines > res_newlines:
                     res += self.environment.newline_sequence * (data_newlines - res_newlines)
+                    if unsafe:
+                        res = wrap_var(res)
             return res
         except (UndefinedError, AnsibleUndefinedVariable) as e:
             if fail_on_undefined:
